import pandas as pd
import numpy as np
from prince import PCA, MCA, FAMD
from sklearn.ensemble import ExtraTreesRegressor, ExtraTreesClassifier
from sklearn.model_selection import RandomizedSearchCV, train_test_split
from sklearn.metrics import r2_score

# *******************************************************

def Trans_data_to_pandas(data, n_x, dims, problem_type):

    if problem_type == "Mixed":
        cols = np.arange(1,dims+1)
        cols = cols.astype(str)
        data = pd.DataFrame(data, columns=cols)
        data[cols[n_x:]] = data[cols[n_x:]].astype(int)
    else:
        data = pd.DataFrame(data)

    return data

# *******************************************************

def Reduce(data, n_x, dims, problem_type, reducer):
    
    data_pd = Trans_data_to_pandas(data, n_x, dims, problem_type)
    
    return np.array(reducer.transform(data_pd))

# *******************************************************

def Inverse(data, inverter_transform, inverter):

    def Back_projection(data, model):

        predictions = {}
        for col, mod in model.items():
            predictions[col] = mod.predict(data)
        
        return predictions
    
    data = pd.DataFrame(data)
    if inverter_transform == "Yes":
        data_reconstructed = inverter.inverse_transform(data)
    elif inverter_transform == "No":
        predictions = Back_projection(data, inverter)
        data_reconstructed = np.vstack(([predictions[i] for i in range(len(inverter))])).T

    return data_reconstructed

# *******************************************************

def Train_reducer(data, n_x, dims, problem_type, n_components, reducer):

    data_pd = Trans_data_to_pandas(data, n_x, dims, problem_type)
    if reducer.__module__ == 'prince.pca':
        reducer_train = PCA()
    elif reducer.__module__ == 'prince.famd':
        reducer_train = FAMD()
    elif reducer.__module__ == 'prince.mca':
        reducer_train = MCA()
    else:
        reducer_train = reducer

    reducer_train.n_components = n_components
    data_red = np.array(reducer_train.fit_transform(data_pd))
    
    return data_red, reducer_train

# *******************************************************

def Train_inverter(data, data_reduced, dims, models):

    for i in range(dims):
        model = models[i]
        model.fit(data_reduced, data[:,i])
        models[i] = model

    return models

# *******************************************************

def Find_reducer(data, n_x, dims, n_components, problem_type):

    def Find_n_components_PCA(x_train, x_test, n, reducer, error_min = 0.5, n_max = 6):

        def Train_Red_comp(x_train, n_trial, reducer):

            reducer.n_components = n_trial
            reducer = reducer.fit(x_train)
            
            return reducer
        
        def Reconstruction_error(x_test, reducer):

            X_test_red = reducer.transform(x_test)
            X_test_reconstructed = reducer.inverse_transform(X_test_red)
            
            return r2_score(x_test, X_test_reconstructed)

        for n in range(1, n_max):
            reducer = Train_Red_comp(x_train, n, reducer)
            error = Reconstruction_error(x_test, reducer)
            if error >= error_min:
                break

        return n
    
    def Find_n_components(data, n, reducer, var_max = 50, n_max = 6):

        def Red(reducer, data, n_trial):

            reducer.n_components = n_trial
            reducer = reducer.fit(data)

            return reducer.cumulative_percentage_of_variance_[-1]
        
        for n in range(1, n_max):
            variance = Red(reducer, data, n)
            if variance >= var_max:
                break

        return n

    # Main program
    
    if problem_type == "Continuous":
        reducer = PCA()
        reducer_search = PCA()
        reducer_train = PCA()
    else:
        if problem_type == "Mixed":
            reducer = FAMD()
            reducer_train = FAMD()
        elif problem_type == "Discrete":
            reducer = MCA()
            reducer_train = MCA()
        else:
            pass

    if n_components is None:
        if reducer is None:
            pass
        else:
            data_pd = Trans_data_to_pandas(data, n_x, dims, problem_type)
            if problem_type == "Continuous":
                x_train, x_test = train_test_split(data_pd)
                dims_red = Find_n_components_PCA(x_train, x_test, dims, reducer_search)
                reducer_train = PCA(n_components=dims_red)
                reducer_train = reducer_train.fit(data_pd)
            else:
                dims_red = Find_n_components(data_pd, dims, reducer_train)
    else:
        dims_red = n_components

    return reducer, reducer_train, dims_red

# *******************************************************

def Find_inverter(data, data_reduced, n_x, n_y, dims, problem_type):

    def Find_best_models(x, z, dims, problem_type, method):

        param_clas = {
            'n_estimators': [50, 100, 200],
            'max_depth': [None, 10, 20, 30],
            'min_samples_split': [2, 5, 10],
            'min_samples_leaf': [1, 2, 4],
            'bootstrap': [True, False]
        }

        if problem_type == "Discrete" and dims < 5:
            random_search = RandomizedSearchCV(method(), param_distributions=param_clas, n_iter=15, cv=dims)
        elif problem_type == "Mixed" and n_y < 5:
            random_search = RandomizedSearchCV(method(), param_distributions=param_clas, n_iter=15, cv=n_y)
        else:
            random_search = RandomizedSearchCV(method(), param_distributions=param_clas, n_iter=15, cv=5)
        random_search.fit(x, z)
        best_params = random_search.best_params_
        best_model = method(**best_params)
        
        return best_model

    def Train_models(x, z, method):

        models = {}
        for col in z.columns:
            model = Find_best_models(x, z[col], problem_type, dims, method)
            model.warm_start = True
            models[col] = model

        return models
    
    data_pd = Trans_data_to_pandas(data, n_x, dims, problem_type)
    if problem_type == "Discrete":
        model = Train_models(data_reduced, data_pd, ExtraTreesClassifier)
    elif problem_type == "Continuous":
        model = Train_models(data_reduced, data_pd, ExtraTreesRegressor)
    elif problem_type == "Mixed":
        model_1 = Train_models(data_reduced, data_pd.iloc[:,:n_x], ExtraTreesRegressor)
        model_2 = Train_models(data_reduced, data_pd.iloc[:,n_x:], ExtraTreesClassifier)
        model = {**model_1, **model_2}
        model_names = np.arange(len(model))
        model = dict(zip(model_names, list(model.values())))

    return model

# *******************************************************

def Red_bounds(mesh, x_l, x_u, n_x, dims, dims_red, problem_type, reducer):

    if reducer is None:
        x_l_red = x_l
        x_u_red = x_u
    else:
        x_mesh_red = Reduce(mesh, n_x, dims, problem_type, reducer)
        x_l_red = np.array([x_mesh_red[0,i] for i in range(dims_red)])
        x_u_red = np.array([x_mesh_red[-1,i] for i in range(dims_red)])

    return x_l_red, x_u_red


import numpy as np
from scipy.stats import norm
from sklearn.preprocessing import StandardScaler

# *******************************************************

def UCB(x, xi, model):
    # Predict the model
    mean, std = model.predict(x)
    af = mean + xi*std
    
    return af

# *******************************************************

def PI(x, x_best, xi, model):

    mean, std = model.predict(x)
    with np.errstate(divide='warn'):
        imp = mean - x_best - xi
        Z = imp / std
        sld = StandardScaler().fit_transform(Z)
        af = norm.cdf(sld)
        af[std == 0.0] = 0.0

    return af

# *******************************************************

def EI(x, x_best, xi, model):
    
    mean, std = model.predict(x)
    with np.errstate(divide='warn'):
        imp = mean - x_best - xi
        Z = imp / std
        sld = StandardScaler().fit_transform(Z)
        af = imp * norm.cdf(sld) + std * norm.pdf(sld)
        af[std == 0.0] = 0.0
    
    return af

# *******************************************************

def PoF(x, models):
    
    pof = []

    for model in models:
        mean, std = model.predict(x)
        std[(std < 1e-100) & (std > 0)] = 1e-99
        Z = -mean/std
        sld = StandardScaler().fit_transform(Z)
        pof.append(norm.cdf(sld))

    return np.prod(np.array(pof), axis=0)

# *******************************************************

def Prob_GPC(x, model):

    gpc, _ = model.predict(x)

    return gpc

# *******************************************************

def AF(x, params, constraints_method, model, models_const):

    xi, _, x_best, AF_name = params.values()

    if AF_name == 'UCB':
        score = UCB(x, xi, model)
    elif AF_name == 'PI':
        score = PI(x, xi, x_best, model)
    elif AF_name == 'EI':
        score = EI(x, xi, x_best, model)
    else:
        pass

    if models_const is None:
        pass
    else:
        if constraints_method == "PoF":
            score_const = PoF(x, models_const)
        elif constraints_method == "GPC":
            score_const = Prob_GPC(x, models_const)
        else:
            pass
        score = score.reshape(-1,1)
        score_const = score_const.reshape(-1,1)
        score = score*score_const

    return score



import numpy as np
import sympy as sp
import multiprocess as mp
import GPy
import warnings
from ..utils.Initialize import Space, Problem_type, Get_constraints, Times_fun, Points_initial_design, x_Generator, Bounds, Points_mesh, AF_params
from ..utils.Dimension_reduction import Reduce, Inverse, Train_reducer, Train_inverter, Find_reducer, Find_inverter, Red_bounds
from ..utils.Models import Select_model, Train_model, Train_models_const, Kernel_discovery
from ..utils.Acq_fun import AF
from ..utils.Set_level_filtration import Slf
from ..utils.Querry_points import Querry
from ..utils.Update import Up_mesh
from ..utils.Aux import Errors, Data_eval, Eval_fun, Eval_const, Regret, Print_results, Create_results

class BO():

    def __init__(self, function, domain, sense,
                 surrogate = "GP",
                 acquisition_function = "UCB",
                 xi = 2,
                 xi_decay = "yes",
                 kernel = None,
                 kern_discovery = "yes",
                 kern_discovery_evals = 2,
                 x_0 = None, 
                 z_0 = None,
                 design = "LHS",
                 p_design = None,
                 parallelization = "no",
                 max_iter = 100, 
                 n_restarts = 5,
                 constraints = None,
                 constraints_method = "PoF",
                 reducer = None,
                 n_components = None,
                 inverter_transform = "No",
                 verbose = 0,
                 ):

        self.function = function
        self.domain = domain
        self.constraints = constraints
        self.x_0 = x_0
        self.z_0 = z_0
        self.sense = sense
        self.surrogate = surrogate
        self.acquisition_function = acquisition_function
        self.kernel = kernel
        self.parallelization = parallelization
        self.kern_discovery = kern_discovery
        self.design = design
        self.p_design = p_design
        self.max_iter = max_iter
        self.n_restarts = n_restarts
        self.kern_discovery_evals = kern_discovery_evals
        self.xi = xi
        self.xi_decay = xi_decay
        self.constraints_method = constraints_method
        self.reducer = reducer
        self.n_components = n_components
        self.inverter_transform = inverter_transform
        self.verbose = verbose

    def optimize(self):

        # *************** Main program ********************
        # Ignore warnings
        warnings.filterwarnings("ignore")
        #sys.setrecursionlimit(10000)
        #
        dims, n_x, n_d, n_c, n_y, x_l, x_u, y_v, dis_val, cat_val, enc_cat, names = Space(self.domain)
        Errors(n_x, n_y, self.design, self.surrogate, self.constraints_method, self.acquisition_function)
        problem_type = Problem_type(n_x, n_y)
        # 
        if self.constraints is None:
            const, n_const = None, None
        else:
            const, n_const = Get_constraints(self.constraints, self.constraints_method)
        # If an initial database in not provided, create using the design of experiments
        if self.x_0 is None:
            # Evaluate an arbitrary point to determine the computation time of the function
            x_trial = x_Generator(x_l, x_u, y_v, n_x, n_y, dims, 1, problem_type, "random")
            x_eval = Data_eval(x_trial, n_c, dims, enc_cat)
            times, z_trial = Times_fun(self.function, x_eval)
            if self.p_design == None:
                self.p_design = Points_initial_design(times, dims, self.design)
            self.x_0 = x_Generator(x_l, x_u, y_v, n_x, n_y, dims, self.p_design, problem_type, self.design)
            x_eval = Data_eval(self.x_0, n_c, dims, enc_cat)
            self.z_0 = self.function(x_eval).reshape(-1,1)
            x, z = np.vstack((self.x_0, x_trial)), np.vstack((self.z_0, z_trial))
        else:
            x = self.x_0
            if self.z_0 is None:
                x_eval = Data_eval(self.x_0, n_c, dims, enc_cat)
                z = self.function(x_eval).reshape(-1,1)
            else:
                self.p_design = len(self.z_0)
                z = self.z_0
        # Generate mesh, or grid
        p_mesh = Points_mesh(dims)
        if p_mesh > 3: 
            mesh = x_Generator(x_l, x_u, y_v, n_x, n_y, dims, p_mesh, problem_type, "Mesh")
        else:
            mesh = x_Generator(x_l, x_u, y_v, n_x, n_y, dims, 1024, problem_type, "Sobol")
        # Find reducer and perform dimensionality reduction
        if self.reducer is None:
            if (problem_type == "Continuous") and (n_x < 5):
                self.reducer = "No"
                reducer_trained = None
                dims_red = dims
                x_red = x
            else:
                self.reducer, reducer_trained, dims_red = Find_reducer(x, n_x, dims, self.n_components, problem_type)
                x_red = Reduce(x, n_x, dims, problem_type, reducer_trained)
        elif self.reducer == "No":
            reducer_trained = None
            dims_red = dims
            x_red = x
        else:
            dims_red = self.n_components
            x_red, reducer_trained = Train_reducer(x, n_x, dims, problem_type, dims_red, self.reducer)
        # Reduce bounds
        x_l_red, x_u_red = Red_bounds(mesh, x_l, x_u, n_x, dims, dims_red, problem_type, reducer_trained)
        # Find inverter
        if self.inverter_transform == "Yes":
            if self.reducer.__module__ == 'prince.mca' or self.reducer.__module__ == 'prince.famd':
                raise ValueError("Reducer module has no inverse_transform")
            else:
                inverter = reducer_trained
        elif self.inverter_transform == "No":
            if self.reducer == "No":
                inverter = None
            else:
                inverter = Find_inverter(x, x_red, n_x, n_y, dims, problem_type)
                inverter = Train_inverter(x, x_red, dims, inverter)
        else:
            pass
        # Generate mesh in the reduced space
        p_mesh_red = Points_mesh(dims_red)
        mesh_red = x_Generator(x_l_red, x_u_red, 0, 0, 0, dims_red, p_mesh_red, "Continuous", "Mesh")
        p_mesh_0 = p_mesh_red
        p_mesh_red, mesh_red = [p_mesh_red], [mesh_red]
        # Tuple of the bounds
        bounds = Bounds(x_l_red, x_u_red, dims_red)
        # Spawning processes
        if self.parallelization == "yes":
            jobs = mp.cpu_count()
        else:
            jobs = 1
        # Select covariance
        if self.kern_discovery == "yes":
            model = Kernel_discovery(x_red, z, dims_red, self.surrogate, self.kern_discovery_evals)
            kernel = model.kern
        elif self.kern_discovery == "no" and self.kernel is None:
            kernel = GPy.kern.RBF(input_dim=dims_red, variance=1.0, lengthscale=1.0)
        else:
            kernel = self.kernel
        # Eval and train constraints
        if self.constraints is None:
            models_const = None
        else:
            g_hat = Eval_const(x, const, n_const, self.constraints_method)
            models_const = Train_models_const(x, g_hat, n_const, self.constraints_method)
        # Initialize parameters of acquisition function
        af_params = AF_params(z, self.xi, self.xi_decay, self.max_iter, self.acquisition_function, self.sense)
        #
        x_symb = sp.Matrix(sp.symbols('x:' + str(dims_red)))
        x_symb_names = sp.Matrix(sp.symbols('x:' + str(dims)))
        q0 = 25
        qf = 75
        delta_q = (qf-q0)/self.max_iter
        q_inc = q0
        q = q0
        flag = 0
        rt = []
        #
        if self.sense == "maximize":
            ix_best = np.argmax(z)
            z_best = np.max(z)
        elif self.sense == "minimize":
            ix_best = np.argmin(z)
            z_best = np.min(z)
        x_best = x[ix_best]
        af_params["f_best"] = z_best
        # Print results if vervose is active
        if self.verbose == 1:
            if names is None:
                header = 'ite  ' +  '  f      ' + str(x_symb_names[0])
            else:
                header = 'ite  ' +  '  f      ' + str(names[0])
            for i in range(1, dims):
                if names is None:
                    header += '      ' + str(x_symb_names[i])
                else:
                    header += '      ' + str(names[i])
            print(header)
            x_best_eval = Data_eval(x_best.reshape(1,-1), n_c, dims, enc_cat)
            x_print = Print_results(x_best_eval, enc_cat)
            z_print = "%.5f" % z_best if 1e-3 < abs(z_best) < 1e3 else "%0.1e" % z_best
            print(0, ' ', z_print, *x_print)
        # Cycle
        for ite in range(1, self.max_iter+1):
            model = Select_model(x_red, z, kernel, self.surrogate)
            # Train model
            model = Train_model(model, self.n_restarts)
            if self.constraints is None:
                models_const = None
            else:
                x_eval = Data_eval(x, n_c, dims, enc_cat)
                g_hat = Eval_const(x_eval, const, n_const, self.constraints_method)
                models_const = Train_models_const(x_red, g_hat, n_const, self.constraints_method)
            # Find conected elements
            connected_elements, n_elements = Slf(mesh_red, p_mesh_red, dims_red, q, jobs, af_params, self.constraints_method, self.sense, model, models_const, AF)
            # Querry point
            x_red_new = Querry(x_symb, connected_elements, n_elements, bounds, dims_red, af_params, self.constraints_method, self.sense, model, models_const, AF)
            #
            if self.reducer == "No":
                x_new = x_red_new
            else:
                x_new = Inverse(x_red_new, self.inverter_transform, inverter)
            #
            x_eval = Data_eval(x_new, n_c, dims, enc_cat)
            z_new = Eval_fun(x_eval, n_elements, self.parallelization, jobs, self.function)
            #
            if jobs == 1:
                z_new_best = z_new
            else:
                if self.sense == "maximize":
                    z_new_best = np.max(z_new)
                elif self.sense == "minimize":
                    z_new_best = np.min(z_new)
                else:
                    pass
            # Update train data
            x, z = np.vstack((x, x_new)), np.vstack((z, z_new))
            if self.reducer == "No":
                x_red = np.vstack((x_red, x_red_new))
            else:
                x_red, reducer_trained = Train_reducer(x, n_x, dims, problem_type, dims_red, self.reducer)
                if self.inverter_transform == "No":
                    inverter = Train_inverter(x, x_red, dims, inverter)
                elif self.inverter_transform == "Yes":
                    inverter = reducer_trained
                else:
                    pass
            # 
            if self.sense == "maximize":
                if z_new_best > z_best:
                    flag = 1
            elif self.sense == "minimize":
                if z_new_best < z_best:
                    flag = 1
            # Update parameters
            if flag == 1:
                if self.reducer != "No":
                    connected_elements_inv = [Inverse(connected_elements[i], self.inverter_transform, inverter) for i in range(n_elements)]
                    connected_elements = [Reduce(connected_elements_inv[i], n_x, dims, problem_type, reducer_trained) for i in range(n_elements)]
                mesh_red, p_mesh_red = Up_mesh(connected_elements, n_elements, p_mesh_0, dims_red)
                af_params["xi"] *= af_params["xi_decay"]
                q_inc += delta_q
                q = int(round(q_inc/5.0)*5.0)
            rt.append(Regret(z_new, np.array(x_red_new), n_elements, model))
            # Print results
            if self.sense == "maximize":
                ix_best = np.argmax(z)
                z_best = np.max(z)
            elif self.sense == "minimize":
                ix_best = np.argmin(z)
                z_best = np.min(z)
            x_best = x[ix_best]
            af_params["f_best"] = z_best
            flag = 0
            if self.verbose == 1:
                x_best_eval = Data_eval(x_best.reshape(1,-1), n_c, dims, enc_cat)
                x_print = Print_results(x_best_eval, enc_cat)
                z_print = "%.5f" % z_best if 1e-3 < abs(z_best) < 1e3 else "%0.1e" % z_best
                print(ite, ' ', z_print, *x_print)

        res = Create_results(x_best, z_best, x, z, x_l, x_u, dims, self.max_iter, self.p_design, self.design, af_params, self.constraints_method, rt, models_const, model)

        return res


from types import SimpleNamespace
import multiprocess as mp
import numpy as np

# *******************************************************

def Errors(n_x, n_y, design_type, surrogate, constraints_method, AF_name):

    valid_var_type = ["continuous", "discrete", "categorical"]
    valid_design_type = ["random", "LHS", "Sobol", "Halton", "Mesh"]
    valid_surrogate_name = ["GP", "SGP"]
    valid_constraints_method = ["PoF", "GPC"]
    valid_af_names = ['UCB', 'EI', 'PI']

    if n_x <= 0 and n_y <= 0:
        raise ValueError("Not valid variable type, valid names are:", *valid_var_type)

    if design_type not in valid_design_type:
        raise ValueError("Not valid design type, valid names are:", *valid_design_type)
    
    if surrogate not in valid_surrogate_name:
        raise ValueError("Not valid acquisition function name, valid names are:", *valid_surrogate_name)
    
    if constraints_method not in valid_constraints_method:
        raise ValueError("Not valid method for constraints, valid names are:", *valid_constraints_method)

    if AF_name not in valid_af_names:
        raise ValueError("Not valid acquisition function name, valid names are:", *valid_af_names)
    
    return None

# *******************************************************

def Data_eval(x, n_c, dims, enc_cat):

    if enc_cat != None:
        x_conv = []
        for i in range(len(enc_cat)):
            ix_cat = (dims - n_c)
            x_conv.append(enc_cat[i].inverse_transform(x[:,ix_cat+i].reshape(-1,1)))
        x_conv = np.array(x_conv).reshape(-1,1)
        x_eval = np.hstack((x[:, :ix_cat], np.asarray(x_conv,object)))
    else:
        x_eval = x
    
    return x_eval

# *******************************************************

def Eval_fun(x, n_elements, parallelization, jobs, function):

    if parallelization == "no":
        x_new = np.array(x).reshape(1,-1)
        z_new = function(x_new)
    elif parallelization == "yes":
        x_new = [x[i].reshape(1,-1) for i in range(n_elements)]
        with mp.Pool(jobs) as pool:
            z_new = pool.map(function, x_new)
        
    return z_new

# *******************************************************

def Eval_const(x, const, n_const, constraints_method):
    
    if constraints_method == "PoF":
        data = [eval(const[i], None, {"x": x}) for i in range(n_const)]
    elif constraints_method == "GPC":
        data = [eval(const[i]['constraint'], None, {"x": x}) for i in range(n_const)]

    return np.array(data).reshape(-1, n_const)

# *******************************************************

def Regret(z_true, x, n_elements, model):
    # Return an average of the reward
    z_pred, _ = model.predict(x)
    rt = [(z_true[i] - z_pred[i]) for i in range(n_elements)]
    rt = sum(rt)

    return rt/n_elements

# *******************************************************

def Print_results(x, enc_cat):

    if enc_cat is None:
        x = x.reshape(-1)
        x_print = ["%.5f" % value if 1e-3 < abs(value) < 1e3 else "%0.1e" % value for value in x]
    else:
        x = x[0]
        x_print = [] 
        for value in x:
            if type(value) == str:
                x_print.append(value)
            elif type(value) == float:
                if 1e-3 < abs(value) < 1e3:
                    x_print.append("%.5f" % value)
                else:
                    x_print.append("%0.1e" % value)
        
    return x_print

# *******************************************************

def Create_results(x_best, z_best, x, z, x_l, x_u, dims, max_iter, points, design, af_params, constraints_method, rt, models_const, model):

    res = {'x_best': x_best, 'f_best': z_best, 
           'x_init': x[0:points], 'f_init': z[0:points], 
           'x_iters': x[points:-1], 'f_iters': z[points:-1],
           'x_l': x_l, 'x_u': x_u, 
           'dims': dims,
           'iters': max_iter, 
           'initial_design': design,
           'initial_points': points,
           'xi': af_params['xi'], 
           'acquisition_function': af_params['AF_name'],
           'regret': rt,
           'constraint_method': constraints_method,
           'models_constraints': models_const,
           'model': model}
    
    return SimpleNamespace(**res)


from .Acq_fun import UCB, PI, EI, PoF, Prob_GPC, AF
from .Aux import Errors, Data_eval, Eval_fun, Eval_const, Regret, Print_results, Create_results
from .Dimension_reduction import Trans_data_to_pandas, Reduce, Inverse, Train_reducer, Train_inverter, Find_reducer, Find_inverter, Red_bounds
from .Initialize import Space, Problem_type, Get_constraints, Times_fun, Points_initial_design, x_Generator, Bounds, Points_mesh, AF_params
from .Models import Select_model, Train_model, Train_models_const, Kernel_discovery
from .Querry_points import Querry
from .Set_level_filtration import Slf
from .Update import Up_mesh


import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from ..utils.Acq_fun import AF

# *******************************************************

def Data_plot(x_l, x_u, dims, n_plot):

    lists = [np.linspace(x_l[i], x_u[i], n_plot) for i in range(dims)]
    # Create a meshgrid data
    x_plot = np.meshgrid(*lists)
    x_plot = np.array(x_plot).T.reshape(-1, dims)
    
    return x_plot

# *******************************************************

def Plot_surrogate(args):

    _, _, x_init, f_init, x_iters, f_iters, x_l, x_u, dims, _, _, initial_points, _, _, _, _, _, model = args.__dict__.values()
    # x_best, f_best, x_init, f_init, x_iters, f_iters, x_l, x_u, dims, iters, initial_design, initial_points, xi, acquisition_function, regret, constraint_method, models_constraints, model

    if dims == 1:
        
        #af_params = {'xi': xi, 'xi_decay': None, 'f_best': f_best, 'AF_name': acquisition_function}
        fig = plt.figure()
        n_plot = 1000
        x_plot = Data_plot(x_l, x_u, dims, n_plot)
        f_mean, f_std = model.predict(x_plot)

        plt.plot(x_plot, f_mean)
        plt.fill_between(x_plot.reshape(-1), (f_mean - f_std).reshape(-1), (f_mean + f_std).reshape(-1), alpha=0.2)
        plt.scatter(x_init, f_init, c='k', s=10, label='Training data')
        plt.scatter(x_iters, f_iters, c='r', s=10, label='Observations')
        plt.xlabel('$x$')
        plt.ylabel('$f(x)$')
        plt.title('Surrogate model')
        plt.legend(bbox_to_anchor=(1.03, 1), loc='upper left', borderaxespad=0.)
        
    elif dims == 2:

        fig, ax = plt.subplots(1, 2, figsize=(15, 5))
        n_plot = 100
        x_plot = Data_plot(x_l, x_u, dims, n_plot)
        f_mean, f_std = model.predict(x_plot)
        x1_plot, x2_plot = x_plot[:,0].reshape(n_plot, n_plot), x_plot[:,1].reshape(n_plot, n_plot)
        x1_init, x2_init = x_init[:,0], x_init[:,1]
        x1_iters, x2_iters = x_iters[:,0], x_iters[:,1]
        f_mean, f_std = f_mean.reshape(n_plot, n_plot), f_std.reshape(n_plot, n_plot)

        p1 = ax[0].contourf(x1_plot, x2_plot, f_mean, cmap='jet')
        ax[0].scatter(x1_init, x2_init, c='k', label='Training data')
        ax[0].scatter(x1_iters, x2_iters, c='r', label='Observations')
        ax[0].set_xlabel('$x_1$')
        ax[0].set_ylabel('$x_2$')
        ax[0].set_title('Surrogate model \n mean')

        p2 = ax[1].contourf(x1_plot, x2_plot, f_std, cmap='jet')
        ax[1].scatter(x1_init, x2_init, c='k', label='Training data')
        ax[1].scatter(x1_iters, x2_iters, c='r', label='Observations')
        ax[1].set_xlabel('$x_1$')
        ax[1].set_ylabel('$x_2$')
        ax[1].set_title('Surrogate model \n uncertanty')
        
        plt.colorbar(p1)
        plt.colorbar(p2)
        plt.legend(bbox_to_anchor=(1.3, 1), loc='upper left', borderaxespad=0.)

    elif dims > 2:

        print('not posible to plot')
        fig = None

    return fig

# *******************************************************

def Plot_AF(args):

    _, f_best, _, _, _, _, x_l, x_u, dims, _, _, _, xi, acquisition_function, _, constraints_method, models_const, model = args.__dict__.values()

    if dims == 1:
        
        af_params = {'xi': xi, 'xi_decay': None, 'f_best': f_best, 'AF_name': acquisition_function}
        fig = plt.figure()
        n_plot = 500
        x_plot = Data_plot(x_l, x_u, dims, n_plot)
        z_acq = AF(x_plot, af_params, constraints_method, model, models_const)

        plt.plot(x_plot, z_acq)
        plt.xlabel('$x$')
        plt.ylabel('$AF(x)$')
        plt.title('Acquisition function')
        
    elif dims == 2:

        af_params = {'xi': xi, 'xi_decay': None, 'f_best': f_best, 'AF_name': acquisition_function}
        fig = plt.figure()
        n_plot = 100
        x_plot = Data_plot(x_l, x_u, dims, n_plot)
        z_acq = AF(x_plot, af_params, constraints_method, model, models_const)
        x1_plot, x2_plot = x_plot[:,0].reshape(n_plot, n_plot), x_plot[:,1].reshape(n_plot, n_plot)
        z_acq = z_acq.reshape(n_plot, n_plot)        

        plt.contourf(x1_plot, x2_plot, z_acq, cmap='jet')
        plt.xlabel('$x_1$')
        plt.ylabel('$x_2$')
        plt.title('Acquisition function')

    elif dims > 2:

        print('not posible to plot')
        fig = None

    return fig

def Plot_regret(args):

    _, _, _, _, _, _, _, _, _, _, _, _, _, _, regret, _, _, _ = args.__dict__.values()
    R = np.cumsum(regret)

    fig = plt.figure()

    plt.plot(regret, label="instantaneous regret")
    plt.plot(R, label="cummulative regret")
    plt.xlabel('Iterations')
    plt.ylabel('R')
    plt.title('Regret')
    plt.legend(bbox_to_anchor=(1.03, 1), loc='upper left', borderaxespad=0.)

    return fig


# Class to find the optimal points of each of the conected components, and the constrains using the Mahalanobis distance

# *******************************************************
# Import libraries

import numpy as np
import sympy as sp
from scipy.optimize import minimize

# *******************************************************

def Querry(x, x_matrix, n_elements, bounds, dims, af_params, constraints_method, sense, model, models_const, Acq_fun):

    # Acquisition function for the optimization
    def Acquisition_function_opt(x, af_params, sense, model, models_const):

        if sense == "maximize":
            af_opt = -Acq_fun(x.reshape(1,-1), af_params, constraints_method, model, models_const)
        elif sense == "minimize":
            af_opt = Acq_fun(x.reshape(1,-1), af_params, constraints_method, model, models_const)
        
        return af_opt

    def Mahalanobis_distance(matrix, n_elements):

        mu = [np.mean(matrix[i], axis=0).reshape(-1, 1) for i in range(n_elements)]
        P = []
        chi = []
        Sigma = []

        if dims == 1:
            for i in range(n_elements):
                var = np.var(matrix[i]/np.sqrt(len(matrix[i])), axis=0)
                if var == 0:
                    Sigma.append(1e-6)
                    P.append(1/(1e-6))
                    chi.append(1.4)
                else:
                    Sigma.append(var)
                    var_inv = 1/var
                    P.append(var_inv)
                    y = matrix[i] - mu[i]
                    dist = [(y[j]*np.sqrt(var_inv)) for j in range(len(y))]
                    chi.append(np.percentile(dist, 95))
        else:                
            for i in range(n_elements):
                if (len(matrix[i]) < 3):
                    cov = np.eye(dims)
                    Sigma.append(cov)
                    P.append(cov)
                    chi.append(1.4)
                else:
                    # Covariance matrix
                    cov = np.cov(matrix[i].T)
                    # Check if the covariance matrix is singular, if so, add a small value to the diagonal
                    if np.linalg.det(cov) == 0:
                        cov = cov + 1e-6*np.eye(dims)
                        # Inverse of the covariance matrix
                    cov_inv = np.linalg.inv(cov)
                    y = matrix[i]-mu[i].T
                    dist = [np.sqrt(y[j].T@cov_inv@y[j]) for j in range(len(y))]
                    Sigma.append(cov_inv)
                    P.append(cov)
                    chi.append(np.percentile(dist, 95))

        return mu, Sigma, P, chi

    def Constrains(x, n_elements, mu, P, chi):

        const = []

        for i in range(n_elements):
            y = (x-mu[i])
            dist = y.T*P[i]*y
            const.append(sp.LessThan(dist[0], chi[i]))
            
        const_lamb = [sp.lambdify(x, (const[i].rhs - const[i].lhs), 'numpy') for i in range(n_elements)]
        # lambda function for the constrains
        def Constraint_lambify(fun):

            lam = lambda x: fun(*x)
            return lam

        constraints = [Constraint_lambify(const_lamb[i]) for i in range(n_elements)]
        
        return constraints
    
    # Main program
    # Check if the initial database is empty
    if x_matrix is None:
        raise ValueError('Initial database is empty')
    # Compute the confidence region
    mu, Sigma, P, chi = Mahalanobis_distance(x_matrix, n_elements)
    # Define the constrains
    constraints = Constrains(x, n_elements, mu, Sigma, chi)
    # Define the optimization problem
    x_opt = []
    for i in range(n_elements):
        res = minimize(Acquisition_function_opt, x0=mu[i], args=(af_params, sense, model, models_const), method='SLSQP', bounds=bounds, constraints={'type': 'ineq', 'fun': constraints[i]})
        x_opt.append(res.x)        

    return x_opt


from setuptools import find_packages, setup

setup(
    name='bopt_slf',
    version='1.0.2',
    description='Bayesian optimization for constrained or unconstrained, continuous, discrete or mixed data problems',
    author='Javier Morlet-Espinosa',
    author_email='a00833961@tec.mx',
    packages = find_packages(),
    install_requires=[
        'numpy>=1.23.5',
        'sympy>=1.11.1',
        'pandas>=2.0.3',
        'GPy>=1.10.0',
        'scipy>=1.10.1',
        'scikit-learn>=1.1.3',
        'properscoring>=0.1',
        'prince>=0.12.1',
        'matplotlib>=3.7.3'
        ]
    )


# *******************************************************
# Import libraries

import GPy
import numpy as np
import properscoring as ps
from sklearn.model_selection import train_test_split

# *******************************************************

def Select_model(x, z, kernel, surrogate):

    if surrogate == "GP":
        model = GPy.models.GPRegression(x, z, kernel)
    elif surrogate == "SGP":
        model = GPy.models.SparseGPRegression(x, z, kernel)
    else:
        raise ValueError('Not valid surrogate model')
    
    return model

# *******************************************************

def Train_model(model, n_restarts):

    model.optimize(optimizer='lbfgsb', max_iters=1000, messages=False)
    model.optimize_restarts(num_restarts=n_restarts, verbose=False)

    return model

# *******************************************************

def Train_models_const(x, g, n_const, constraints_method):

    def Train_PoF(x, g, n_const):
        
        models = []
        for i in range(n_const):
            model = GPy.models.GPRegression(x, g[:,i].reshape(-1,1))
            model.optimize()
            models.append(model)

        return models
    
    def Train_GPC(x, g):

        y = (g == True).all(axis=1)
        y = y.astype(int)
        model = GPy.models.GPClassification(x, y.reshape(-1,1))
        model.optimize()

        return model
    
    if constraints_method == "PoF":
        models_const = Train_PoF(x, g, n_const)
    elif constraints_method == "GPC":
        models_const = Train_GPC(x, g)
    else:
        pass

    return models_const

# *******************************************************

def Kernel_discovery(x, z, dims, surrogate, evals, err_min = 0.25):

    def Search(x, z,  kernels):

        # Train the model with the base kernels
        ll, k, models = [], [], {}
        for name, kernel in kernels.items():
            model = Select_model(x, z, kernel, surrogate)
            model.optimize()
            ll.append(model.log_likelihood())
            k.append(model._size_transformed())
            models[name] = model

        return ll, k, x.shape[0], models

    def Bic(ll, k, n): return -2*ll + k*np.log(n)

    def Sort_bic(ll, k, n, models):

        BICs = []
        d = len(models)
        for i in range(d):
            BICs.append(Bic(ll[i], k[i], n))
        
        ordered = [x for _, x in sorted(zip(BICs, models.keys()))]

        return {ordered[0]: models[ordered[0]]}

    x_train, x_test, z_train, z_test = train_test_split(x, z, test_size=0.2)
    # Base kernels: SE, Periodic, Linear, RatQuad  
    kernels = {"linear": GPy.kern.Linear(input_dim=dims),
                    "RBF": GPy.kern.RBF(input_dim=dims, variance=1.0, lengthscale=1.0),
                    "Mater_52": GPy.kern.Matern52(input_dim=dims, variance=1.0, lengthscale=1.0),
                    "Periodic": GPy.kern.StdPeriodic(input_dim=dims, variance=1.0, lengthscale=1.0, period=1.0)
                    }

    for _ in range(evals):

        base_kernels = kernels.copy()
        ll, k, n, models = Search(x_train, z_train, base_kernels)
        best_model = Sort_bic(ll, k, n, models)
        base_model = list(best_model.values())[0]
        mean, var = base_model.predict(x_test)
        error = ps.crps_gaussian(z_test, mean, var).mean()
        # Composing best model with base kernels using addition and multiplication
        kernels = {}
        base_model_name = list(best_model.keys())[0]
        base_model_kern = base_model.kern

        for name, kernel in base_kernels.items():
            kernels[base_model_name + "+" + name] = GPy.kern.Add([base_model_kern, kernel])
            kernels[base_model_name + "*" + name] = GPy.kern.Prod([base_model_kern, kernel])

        if error < err_min:
            break

    return base_model


import numpy as np

def Up_mesh(mesh, n_mesh, p_mesh, dims):

    def New_boundaries(mesh, n_mesh, dims):

        x_l_new, x_u_new = [], []

        for i in range(n_mesh):
            x_l_new.append([np.min(mesh[i][:,j]) for j in range(dims)])
            x_u_new.append([np.max(mesh[i][:,j]) for j in range(dims)])
        
        return x_l_new, x_u_new
    
    def Points_mesh(x_l, x_u, points_mesh, n_mesh, dims):

        l = [[(x_u[i][j] - x_l[i][j]) for i in range(n_mesh)] for j in range(dims)]
        a = [np.prod(l[0][i]) for i in range(n_mesh)]
        pts = a/np.max(a)
        points_mesh_min = int(np.sqrt(points_mesh))

        return [points_mesh_min if int(x*points_mesh) < points_mesh_min else int(x*points_mesh) for x in pts]

    x_l_new, x_u_new = New_boundaries(mesh, n_mesh, dims)
    points_mesh_new = Points_mesh(x_l_new, x_u_new, p_mesh, n_mesh, dims)
    x_mesh_new = []

    for i in range(n_mesh):
        lists = [np.linspace(x_l_new[i][j], x_u_new[i][j], points_mesh_new[i]) for j in range(dims)]
        x_mesh_new.append(np.meshgrid(*lists))
        x_mesh_new[i] = np.array(x_mesh_new[i]).T.reshape(-1, dims)

    return x_mesh_new, points_mesh_new


# *******************************************************
# Import libraries

import numpy as np
from scipy.stats import qmc
from sklearn.preprocessing import OrdinalEncoder
import time

# *******************************************************

def Space(domain):

    """ 
    Gets an array for the lower and upper bounds of the continuous variables, and a tuple for the discrete variables.
    """

    dims = len(domain)
    n_x = 0
    n_d = 0
    n_c = 0
    x_l = []
    x_u = []
    dis_val = []
    cat_val = []
    enc_cat = []
    names = []

    for i in range(dims):
        try:
            names.append(domain[i]['name'])
        except:
            names = 0
        if domain[i]['type'] == "continuous":
            n_x += 1
            x_l.append(domain[i]['domain'][0])
            x_u.append(domain[i]['domain'][1])
        elif domain[i]['type'] == "discrete":
            n_d += 1
            dis_val.append(domain[i]['domain'])
        elif domain[i]['type'] == "categorical":
            n_c += 1
            enc_cat.append(OrdinalEncoder())
            X = domain[i]['domain']
            X = [[(X[i])] for i in range(len(X))]
            X_trans = tuple(enc_cat[n_c-1].fit_transform(X).reshape(-1).astype(int))
            cat_val.append(X_trans)
        else:
            pass
            
    x_l, x_u = np.array(x_l), np.array(x_u)
    n_y = n_d + n_c
    y_v = [item for sublist in [dis_val, cat_val] for item in sublist]
    if n_d == 0:
        dis_val = 0
    if n_c == 0:
        enc_cat = None
        cat_val = 0
    if names == 0:
        names = None

    return dims, n_x, n_d, n_c, n_y, x_l, x_u, y_v, dis_val, cat_val, enc_cat, names

# *******************************************************

def Problem_type(n_x, n_y):

    """ 
    Returns the type of problem depending of the characteristics of the inputs:
    * Continuous if there are only continuous variables
    * Discrete if there are only discrete variables
    * Mixed if there are continuous and discrete variables
    """
    
    if n_x > 0 and n_y == 0:
        problem_type = "Continuous"
    elif n_y > 0 and n_x == 0:
        problem_type = "Discrete"
    elif n_x > 0 and n_y > 0:
        problem_type = "Mixed"
    else:
        pass

    return problem_type

# *******************************************************

def Get_constraints(constraints, constraints_method):

    "Transforms tuple of constraints into list"

    n_const = len(constraints)

    if constraints_method == "PoF":
        symbols = ['<=', '>=']
        const = []
        for i in range(n_const):
            for s in symbols:
                try:
                    index_const = constraints[i]['constraint'].index(s)
                    const.append(constraints[i]['constraint'][:index_const-1])
                except:
                    pass
    elif constraints_method == "GPC":
        const = constraints
    
    return const, n_const

# *******************************************************

def Times_fun(fun, x):

    """ 
    Generates the number of points of the mesh or grid, depending on the cost of the evaluation of the function, and the dimensions of the problem
    """
    
    start = time.time()
    f_eval = fun(x.reshape(1,-1))
    end = time.time()
    
    return (end - start), f_eval

# *******************************************************

def Points_initial_design(times, dims, design, c_param = 50):

    if times <= 1:
        exp_param = 0.25
    else: 
        exp_param = 0.95
    
    points_D = int(c_param - c_param/(1+(1/times)**exp_param))

    if design == "Mesh":
        points_D = int(np.ceil(points_D)**(1/dims))
    elif design == "Sobol":
        points_D = int(np.ceil(np.sqrt(points_D))**2)
    else:
        pass

    if points_D < 3:
        points_D = int(3)

    return points_D

# *******************************************************

def x_Generator(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, design_type):

    def Random_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type):

        def Bounds_y(y_v, n_y):

            y_l = []
            y_u = []
            flag = int(all([(np.diff(y_v[i]) == 1).all() for i in range(n_y)]))
            
            for i in range(n_y):
                if flag == 1:
                    y_l.append(y_v[i][0])
                    y_u.append(y_v[i][-1]+1)
                elif flag == 0:
                    y_l.append(0)
                    y_u.append(1)
                else:
                    pass
            return y_l, y_u, flag

        def X_rand(x_l, x_u, dims, points):

            return np.random.uniform(x_l, x_u, size=(points, dims))
        
        def Y_rand(y_v, n_y, points):

            y_l, y_u, flag = Bounds_y(y_v, n_y)

            if flag == 1:
                y_rand = np.random.randint(y_l, y_u, size=(points, dims))
            elif flag == 0:
                size_y = [len(y_v[i]) for i in range(n_y)]
                l = [np.repeat(1/size_y[i], size_y[i]) for i in range(n_y)]
                sample = [np.random.choice(y_v[i], p=l[i], size=(points, 1)).reshape(-1) for i in range(n_y)]
                y_rand = np.array(sample).T
            else:
                pass
            
            return y_rand

        if problem_type == "Continuous":
            variables = X_rand(x_l, x_u, dims, points)
        elif problem_type == "Mixed":
            x_variables = X_rand(x_l, x_u, n_x, points)
            y_variables = Y_rand(y_v, n_y, points)
            variables = np.hstack((x_variables, y_variables))
        elif problem_type == "Discrete":
            variables = Y_rand(y_v, dims, points)[0]
        else:
            pass

        return variables
    
    def QMC_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, design_type):

        def Bounds_y(y_v, n_y):

            y_l = []
            y_u = []
            flag = int(all([(np.diff(y_v[i]) == 1).all() for i in range(n_y)]))
            
            for i in range(n_y):
                if flag == 1:
                    y_l.append(y_v[i][0])
                    y_u.append(y_v[i][-1])
                elif flag == 0:
                    y_l.append(0)
                    y_u.append(1)
                else:
                    pass
            return y_l, y_u, flag

        def X_qmc(x_l, x_u, points, method):
                    
            sample = method.random(n=points)

            return qmc.scale(sample, x_l, x_u)

        def Y_qmc(y_v, n_y, points, method):

            def Assign_y(y_v, n_y, y_norm, points):

                size_y = [len(y_v[i]) for i in range(n_y)]
                l = [[i/size_y[j] for i in range(size_y[j]+1)] for j in range(n_y)]
                for i in range(len(l)):
                    l[i][-1] = l[i][-1] + 0.0001
                y_new = np.empty([points, n_y])

                for i in range(n_y):
                    for j in range(len(y_norm[:,i])):
                        for k in range(len(l[i])-1):
                            if l[i][k] <= y_norm[j,i] < l[i][k+1]:
                                y_new[j,i] = y_v[i][k]
                                break
                
                return y_new
            
            y_l, y_u, flag = Bounds_y(y_v, n_y)
            if flag == 1:
                y_qmc = method.integers(l_bounds=y_l, u_bounds=y_u, n=points, endpoint=True)
            elif flag == 0:
                y_norm = X_qmc(y_l, y_u, points, method)
                y_qmc = Assign_y(y_v, n_y, y_norm, points)        

            return y_qmc

        def Select_method(design_type, dims):
            
            if design_type == "LHS":
                method = qmc.LatinHypercube(d=dims)
            elif design_type == "Sobol":
                method = qmc.Sobol(d=dims)
            elif design_type == "Halton":
                method = qmc.Halton(d=dims)
            else:
                pass

            return method

        if problem_type == "Continuous":
            method = Select_method(design_type, dims)
            variables = X_qmc(x_l, x_u, points, method)
        elif problem_type == "Mixed":
            method = Select_method(design_type, n_x)
            x_variables = X_qmc(x_l, x_u, points, method)
            method = Select_method(design_type, n_y)
            y_variables = Y_qmc(y_v, n_y, points, method)
            variables = np.hstack((x_variables, y_variables))
        elif problem_type == "Discrete":
            method = Select_method(design_type, n_y)
            variables = Y_qmc(y_v, n_y, points, method)
        else:
            pass

        return variables
    
    def Mesh_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type):

        def X_mesh(lower_bound, upper_bound, dims, points):

            lists = [np.linspace(lower_bound[i], upper_bound[i], points) for i in range(dims)]
            mesh = np.meshgrid(*lists)
            
            return np.array(mesh).T.reshape(-1, dims)
        
        def Bounds_y(n_y):

            y_l = np.zeros(n_y)
            y_u = np.ones(n_y)
            
            return y_l, y_u

        def Assign_y(y_v, n_y, y_norm, points):

                size_y = [len(y_v[i]) for i in range(n_y)]
                l = [[i/size_y[j] for i in range(size_y[j]+1)] for j in range(n_y)]
                for i in range(len(l)):
                    l[i][-1] = l[i][-1] + 0.0001
                y_new = np.empty([points, n_y])
                for i in range(n_y):
                    for j in range(len(y_norm[:,i])):
                        for k in range(len(l[i])-1):
                            if l[i][k] <= y_norm[j,i] < l[i][k+1]:
                                y_new[j,i] = y_v[i][k]
                                break
                
                return y_new

        if problem_type == "Continuous":
            variables = X_mesh(x_l, x_u, dims, points)
        elif problem_type == "Mixed":
            y_l, y_u = Bounds_y(n_y)
            all_vars = X_mesh(np.append(x_l, y_l), np.append(x_u, y_u), dims, points)
            x_variables = all_vars[:, :n_x]
            y_norm = all_vars[:, n_x:]
            y_variables = Assign_y(y_v, n_y, y_norm, y_norm.shape[0])
            variables = np.hstack((x_variables, y_variables))
        elif problem_type == "Discrete":
            y_l, y_u = Bounds_y(n_y)
            y_norm = X_mesh(y_l, y_u, n_y, points)
            variables = Assign_y(y_v, n_y, y_norm, y_norm.shape[0])
        else:
            pass

        return variables

    # Main program
    if design_type == "random":
        variables = Random_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type)
    elif design_type == "LHS":
        variables = QMC_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, "LHS")
    elif design_type == "Sobol":
        points = int(np.ceil(np.sqrt(points))**2)
        variables = QMC_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, "Sobol")
    elif design_type == "Halton":
        variables = QMC_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, "Halton")
    elif design_type == "Mesh":
        variables = Mesh_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type)
    else:
        pass
    
    return variables

# *******************************************************

def Bounds(x_l, x_u, dims):

    bnds = np.sort(np.array([[x_l[i], x_u[i]] for i in range(dims)]))
    return tuple(map(tuple, bnds))

# *******************************************************

def Points_mesh(dims, r1=9):

    """ 
    Generates the points mesh
    """
    points = int(np.ceil(2**r1)**(1/dims))

    if points < 3:
        points = 3

    return points

# *******************************************************

def AF_params(z, xi, xi_decay, iters, AF_name, sense):

    if xi_decay == "yes":
        xi_decay = (1/xi)**(1/iters)
    elif xi_decay == "no":
        xi_decay = 1
    else:
        pass

    if sense == "maximize":
        f_best = np.max(z)
    elif sense == "minimize":
        f_best = np.min(z)
    else:
        pass

    return {'xi': xi, 'xi_decay': xi_decay, 'f_best': f_best, 'AF_name': AF_name}